#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "serjtag.h"
#include "ft245r.h"
#include "jtag.h"

#include "target.h"
#include "tgtlscc.h"
#include "tgtxil.h"

extern char *program;
extern char *progname;
extern int verbose;

void jtag_test(CABLE cbl)
{
	uint32_t r;
	double st,et;

	st = dtime();
	usleep(200000);
	et = dtime();
	printf("usleep 200ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

    if (0 && cbl->delay) {
	st = dtime();
	(cbl->delay)(cbl, 4000);
	et = dtime();
	printf("delay 4ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	(cbl->delay)(cbl, 9000);
	et = dtime();
	printf("delay 9ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	(cbl->delay)(cbl, 20000);
	et = dtime();
	printf("delay 20ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	(cbl->delay)(cbl, 200000);
	et = dtime();
	printf("delay 200ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	(cbl->delay)(cbl, 400000);
	et = dtime();
	printf("delay 400ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	(cbl->delay)(cbl, 600000);
	et = dtime();
	printf("delay 600ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);
    }

	jtag_init(cbl);

	st = dtime();
	jtag_IDLE(cbl, 20, 5);
	et = dtime();
	printf("jtag_IDLE 20ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	jtag_IDLE(cbl, 20, 5);
	et = dtime();
	printf("jtag_IDLE 20ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	st = dtime();
	jtag_IDLE(cbl, 200, 5);
	et = dtime();
	printf("jtag_IDLE 200ms = %7.3f ms\n", (et - st) * 1000.0);
	fflush(stdout);

	r = 0;
//verbose = 256;
/*
 * ! Shift in IDCODE(0x16) instruction
 * SIR     8       TDI  (16);
 * SDR     32      TDI  (FFFFFFFF)
 *                 TDO  (0128D043)
 *                 MASK (FFFFFFFF);
 */
	jtag_SIR32(cbl, LSCC_IDCODE, 8, JTAG_RECV, &r);
	printf ("jtag idcode I = %08x\n" ,r);
fflush(stdout);
	jtag_SDR32(cbl, 0xffffffff, 32, JTAG_RECV, &r);
	printf ("jtag idcode D = %08x\n" ,r);

	jtag_IDLE(cbl, 1, 5);
//verbose = 256;
	jtag_SIR32(cbl, LSCC_IDCODE, 8, JTAG_RECV, &r);
	printf ("jtag idcode I = %08x\n" ,r);
fflush(stdout);
	jtag_SDR32(cbl, 0xffffffff, 32, JTAG_RECV, &r);
	printf ("jtag idcode D = %08x\n" ,r);
verbose = 1;
/* 
 * ! Shift in READ USERCODE(0x17) instruction
 * SIR     8       TDI  (17);
 * RUNTEST IDLE    5 TCK   1.00E-003 SEC;
 * SDR     32      TDI  (FFFFFFFF)
 *                 TDO  (FFFFFFFF)
 *                 MASK (00000000);
 */
	jtag_SIR32(cbl, LSCC_USERCODE, 8, JTAG_RECV, &r);
	printf ("jtag usercode I = %08x\n" ,r);
	jtag_IDLE(cbl, 1, 5);
	jtag_SDR32(cbl, 0xffffffff, 32, JTAG_RECV, &r);
	printf ("jtag usercode D = %08x\n" ,r);
	jtag_IDLE(cbl, 1, 5);
}

void spi_test(TARGET tgt, CABLE cbl)
{
	int i,j;
	uint32_t s,r;
	printf ("spi_test -- sel_chan test\n");
	fflush(stdout);
	for (i=0; i< 4; i++) {
		jtag_reset(cbl);
		(tgt->sel_chan)(tgt, cbl, (i&1)?1:2);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
	}
	printf ("spi_test -- spi_echo test1\n");
	fflush(stdout);
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, SPI_CHAN);
	for (i=0; i< 12; i++) {
		s = (i<10)?((0x20 +i) & 0xff):0xaa;
		if (i==5) s |= 0x5500;
		if (i==6) s |= 0xaa5500;
//if ((i==5) || (i==6)) verbose = 256;
		jtag_SDR32R(cbl, s , (i==6)?24:(i==5)?16:8, JTAG_RECV|JTAG_CONT, &r);
		printf ("  %02x %02x (%02x)\n" , s, r, r<<1);

verbose = 1;
		fflush(stderr);
		fflush(stdout);
		for (j=0; j< 10; j++) {
			usleep(100000);
		}
	}
	jtag_reset(cbl);
	printf ("spi_test done\n");
	fflush(stdout);
}

/* TPI instructions */
#define TPI_SLD_POSTI 0x24 //
#define TPI_SST_POSTI 0x64 //
#define TPI_SETPR_LO  0x68 //
#define TPI_SETPR_HI  0x69 //
#define ISP_TPI_SIN   0x10 //
#define TPI_SOUT      0x90 //

void isp_test(TARGET tgt, CABLE cbl)
{
	int i,j;
	uint32_t s,r;
	printf ("isp_test -- sel_chan \n");
	fflush(stdout);

	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, ISP_CHAN);
	for (j=0; j< 5; j++) {
		usleep(100000);
	}

        s = (TPI_SETPR_LO << 8);
        jtag_SDR32R(cbl, s, 16, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %04x r = %04x (%s)\n" , s, r, "TPI_SETPR_LO");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}

        s = (TPI_SETPR_HI << 8) | 0x40;
        jtag_SDR32R(cbl, s, 16, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %04x r = %04x (%s)\n" , s, r, "TPI_SETPR_HI");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}

	for (i=0; i< 5; i++) {
	    s = ( TPI_SLD_POSTI << 8);
            jtag_SDR32R(cbl, s, 16, JTAG_RECV|JTAG_CONT, &r);
	        printf (" %04x r = %04x (%s)\n" , s, r, "TPI_SLD_POSTI");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
	}
#if 1
        s = TPI_SETPR_LO;
        jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %02x r = %02x (%s)\n" , s, r, "TPI_SETPR_LO");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
        s = 0x55;
        jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %02x r = %02x (%s)\n" , s, r, "2nd byte");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}

        s = TPI_SETPR_HI;
        jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %02x r = %02x (%s)\n" , s, r, "TPI_SETPR_HI");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
        s = 0x40;
        jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	    printf (" %02x r = %02x (%s)\n" , s, r, "2nd byte");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}


	for (i=0; i< 5; i++) {
	    s = TPI_SLD_POSTI;
            jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	        printf (" %02x r = %02x (%s)\n" , s, r, "TPI_SLD_POSTI");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
	    s = 0xaa; // TPI_SLD_POSTI
            jtag_SDR32R(cbl, s, 8, JTAG_RECV|JTAG_CONT, &r);
	        printf (" %02x r = %02x (%s)\n" , s, r, "2nd byte");
		fflush(stdout);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
	}
#endif

	jtag_reset(cbl);
	printf ("isp_test done\n");
	fflush(stdout);
}

void rd1107_test(TARGET tgt, CABLE cbl)
{
	int i,j;
	uint32_t s,r;
	printf ("rd1107_test -- sel_chan test\n");
	fflush(stdout);
	for (i=0; i< 4; i++) {
		jtag_reset(cbl);
		(tgt->sel_chan)(tgt, cbl, (i&1)?1:2);
		for (j=0; j< 5; j++) {
			usleep(100000);
		}
	}
	printf ("rd1107_test -- start \n");
	fflush(stdout);
	jtag_reset(cbl);
	(tgt->sel_chan)(tgt, cbl, SPI_CHAN);

	jtag_SDR32R(cbl, 0x00 , 8, JTAG_RECV, &r); // addr 0: command
		printf (" A00 r = %02x\n" , r);
	jtag_SDR32R(cbl, 0x00 , 8, JTAG_RECV, &r); // addr 0: command
		printf (" A00 r = %02x\n" , r);

	jtag_SDR32R(cbl, 0x00 , 8, JTAG_RECV, &r); // addr 0: command
		printf (" A00 r = %02x\n" , r);
	jtag_SDR32R(cbl, 0x80 , 8, JTAG_RECV, &r); // enable timer
		printf ("  80 r = %02x\n" , r);
	jtag_SDR32R(cbl, 0x01 , 8, JTAG_RECV, &r); // addr 1: frequency
		printf (" A01 r = %02x\n" , r);
	jtag_SDR32R(cbl, 0x01 , 8, JTAG_RECV, &r); // 1KHz
		printf ("  01 r = %02x\n" , r);

	for (i=0; i< 10; i++) {
		jtag_SDR32R(cbl, 0x02 , 8, JTAG_RECV, &r); // addr 2: duty cycle
		printf (" A02 r = %02x\n" , r);
		jtag_SDR32R(cbl, 10*i , 8, JTAG_RECV, &r); // 10*i %
		printf ("  %02x r = %02x\n" , 10*i, r);
		printf ("  duty %d \n" , 10*i);
//verbose = 1;
		fflush(stderr);
		fflush(stdout);
		for (j=0; j< 10; j++) {
			usleep(100000);
		}
	}
	jtag_reset(cbl);
	printf ("rd1107_test done\n");
	fflush(stdout);
}
