#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"

#ifdef STANDALONE_TEST
int verbose;
#else
extern int verbose;
#endif

#define  ERR_FORMAT(n)	(-(n))

int jed_load(FILE *fp, uint8_t *buf, int bufsiz) {
	uint8_t  *p = buf;
	uint8_t  *q = 0;
	int i;
	int len;
	int bit;
	int user_id_valid = 0;
	int feature_len = 0;
	uint32_t user_id = 0;
	uint8_t feature[16];
	int bit_count = 0;
	int bit_size = 0;
	int data_sum = 0;
	int sum = 0;
	int state = 0;
	uint8_t line[256];

	memset(buf, 0, bufsiz);

	while (fgets(line, 256, fp)) {
		len = strlen(line);
		if (line[len-1] == '\n') len--;
		if (line[len-1] == '\r') len--;
		line[len] = 0;
		if (state == 0) {
			if (!strncmp("NOTE ", line, 5)) {
				if (verbose) {
					fprintf(stderr,"%s\n", line);
				}
				continue;
			} else if (!strncmp("QF", line, 2) && (line[len-1]=='*')) {
				sscanf(line+2,"%d", &bit_size);
			} else if ((line[0] == 'C') && (line[len-1]=='*')) {
				sscanf(line+1,"%x", &data_sum);
				if (verbose) {
					fprintf(stderr,"%s (sum: %08x)\n", line, sum);
				}
			} else if ((line[0] == 'E') && ((line[1] == '0') || (line[1] == '1'))) {
				feature_len = 0;
				state = 2;
				q = feature;
				goto proc_feature;
			} else if ((line[0] == 'U') && (line[1] == 'H') 
					&& (line[len-1]=='*')) {
				sscanf(line+2,"%x", &user_id);
				user_id_valid = 1;
			} else if ((line[0] == 'U') && (line[1] == 'A') 
					&& (line[len-1]=='*')) {
			    user_id = 0;
			    for (i=2; i<len -1; i++) {
				user_id <<= 8;
				user_id |= line[i];
			    }
			    user_id_valid = 1;
			} else if ((line[0] == 'U') && (line[len-1]=='*')) {
			    int cnt = 0;
			    user_id = 0;
			    for (i=1; i<len -1; i++) {
				if (line[i] == '0') {
					user_id &= ~(1 << (31 - cnt));
					cnt++;
				} else if (line[i] == '1') {
					user_id |= (1 << (31 - cnt));
					cnt++;
				}
			    }
			    user_id_valid = 1;
			} else if (!strcmp("L0000000", line)) {
				state = 1;
			} else if (!strcmp("L000000", line)) {
				state = 1;
			} else if (line[0] == 'L') {
			    for (i=1; i < len; i++) {
				if ( (line[i] < '0') || (line[i] > '9') ) {
					goto cont;
				}
			    }
			    sscanf(line+1,"%d", &bit);
			    if (verbose) {
				fprintf(stderr,"%s (bit_count: %d sum: %08x)\n", line, bit_count, sum);
			    }
			    if (bit == bit_count) {
				state = 1;
			    }
			}
	cont:
			;
		} else if (state == 1) {
			if (!strcmp("*", line)) {
				state = 0;
			} else {
			    for (i=0; i< len; i++) {
				if ((line[i] == '0') || (line[i] == '1')) {
					bit = bit_count % 8;
					if ((p - buf) < bufsiz) {
					    if (line[i] == '0')
						*p &= ~(1 << bit);
					    else
						*p |= (1 << bit);
					}
					if (bit == 7) {
						sum += *p;
						p++;
					}
					bit_count++;
				} else if (line[i] == '*') {
					state = 0;
				}
			    }
			}
		} else if (state == 2) {
			if (!strcmp("*", line)) {
				state = 0;
			} else {
proc_feature:
			    for (i=0; i< len; i++) {
				if ((line[i] == '0') || (line[i] == '1')) {
					bit = feature_len % 8;
					if ((q - buf) < 128) {
					    if (line[i] == '0')
						*q &= ~(1 << bit);
					    else
						*q |= (1 << bit);
					}
					if (bit == 7) {
						q++;
					}
					feature_len++;
				} else if (line[i] == '*') {
					state = 0;
				}
			    }
			}
		}
	}
	if (feature_len && ((p - buf) + (feature_len/8) <= bufsiz)) {
		q = feature + feature_len/8;
		for (i=0; i< feature_len; i += 8) {
			*p++ = *(--q);
		}
	}
	if (user_id_valid && ((p - buf) + 4 <= bufsiz)) {
		*p++ = user_id & 0xff;
		*p++ = (user_id >> 8) & 0xff;
		*p++ = (user_id >> 16) & 0xff;
		*p++ = (user_id >> 24) & 0xff;
	}
	if (verbose) {
		fprintf(stderr,"bufsiz %d bit_size %d bit_count %d\n",bufsiz, bit_size, bit_count);
		fprintf(stderr,"user_id %08x data_sum %04x sum %04x \n",user_id, data_sum, sum);
		fprintf(stderr,"feature_len %d [", feature_len);
		for (i=0; i< feature_len; i+= 8) 
		    fprintf(stderr,"%02x", feature[i/8]);
		fprintf(stderr,"]\n");
	}
	if (((sum & 0xffff) != data_sum) 
		&& (((sum + 0xf) & 0xffff) != data_sum) )
		return -1;
	if (bit_size != bit_count) return -1;
	return (p - buf);
}

#ifdef STANDALONE_TEST

main(int argc, char *argv[]) {
	FILE *fp;
	int r;
	uint8_t *buf;

	build_rev();
	fp = fopen(argv[1],"rb");
	if (fp == NULL) exit(1);
	buf = malloc(1024 * 1024 * 2);
	r = jed_load(fp, buf, 1024 * 1024 * 2);
	printf("size %d\n", r);
	dump(buf, 0x0000, (r>0)? r : 256, 0);
}
#endif
