`define TARGET_MachXO2
//`define TARGET_MachXO
`define SYS_DRIVEN_CLK4PH
//`define USE_DMY_CLOCK
`define CLK_SEL 1'b0  	// 0: OSCH 1: EXTOSC
//`define EXTOSC_PASSIVE
`define ISP_DEBUG_LED
//`define SPI_DEBUG_LED

`ifdef TARGET_MachXO
   `define JTAG_BSCAN JTAGD
`elsif TARGET_MachXO2
   `define JTAG_BSCAN JTAGF
`endif

module tool_test (
		// JTAG 
		  input TOP_TDI
		, input TOP_TCK
		, input TOP_TMS
		, output  TOP_TDO

		// LED
		,  output [7:0] LED
`ifdef USE_SERIAL
		// Serial
		, input TXD, RTS, DTR
		, inout RXD, CTS
`endif
`ifdef TARGET_MachXO2
		// External OSC
		, input EXTOSC
		, output EXTOSC_EN
`endif
		);

PUR pur(.PUR(1'b1));

  `JTAG_BSCAN bscan (
`ifdef TARGET_MachXO
	.JRST(BSCAN_RESET_N) 
`elsif TARGET_MachXO2
	.JRSTN(BSCAN_RESET) // H : in TestLogicReset State
`endif
	, .JSHIFT(BSCAN_SHIFT)
//	, .JUPDATE(BSCAN_UPDATE)

	, .TDI(TOP_TDI)
	, .TCK(TOP_TCK)
	, .TMS(TOP_TMS) 
	, .TDO(TOP_TDO)

	, .JTDI(JTDI_IN)
	, .JTCK(JTCK_IN) // negedge capt, posedge shift

	// signals for IPA
	, .JRTI1(JRTI1)
	, .JCE1(SEL1)
	, .JTDO1(JTDO1)

	// signals for IPB
	, .JRTI2(JRTI2)
	, .JCE2(SEL2)
	, .JTDO2(JTDO2) 
	);

`ifdef TARGET_MachXO
    wire BSCAN_RESET = ~BSCAN_RESET_N;
`elsif TARGET_MachXO2
    wire BSCAN_RESET_N = ~BSCAN_RESET;
`endif

  wire CLK;
  
//############  SPI/ISP TIMING GENERATION #####################################

  reg [15:0] reset_sft     = 16'h0000;
    always @(negedge CLK)
    begin
        reset_sft <=  { reset_sft[15-1:0] , 1'b1 };
    end
  wire RESET = reset_sft[15];

  reg r_jtck;
  reg r_jtck_prev;
  reg r_cs1;
  reg r_cs1_prev;
  reg r_cs2;
  reg r_cs2_prev;
  
  always @(posedge CLK)
    begin
	r_jtck_prev <= JTCK_IN;
	if (r_jtck_prev == JTCK_IN)
		r_jtck <= ~r_jtck_prev;
	if (r_jtck_prev & r_jtck & (r_jtck_prev == JTCK_IN)) // negedge JTCK
	    begin
		    r_cs1_prev <= ~(SEL1 & BSCAN_SHIFT);
		    r_cs1      <= r_cs1_prev;
		    r_cs2_prev <= ~(SEL2 & BSCAN_SHIFT);
		    r_cs2      <= r_cs2_prev;
	    end
    end
  wire JTCK1 = r_jtck;
  wire JTCK2 = r_jtck;
  wire JTDI  = JTDI_IN;
  wire CS1 = r_cs1;
  wire CS2 = r_cs2;

//############  SPI TEST MODULE #####################################

  wire [7:0] SPI_DATA;
  wire [7:0] SPI_DATA_I;
  wire [1:0] SPI_COUNT;
  reg [7:0] r_data;
  jtag_spi JTAG_SPI (.CLK(CLK)
		, .MOSI(JTDI)
                , .SCK(JTCK1)
                , .CS(CS1)
                , .MISO(JTDO1)
		, .SPI_DATA(SPI_DATA)
		, .SPI_DATA_VALID(SPI_DATA_VALID)
		, .SPI_DATA_REQ(SPI_DATA_REQ)
		, .SPI_DATA_COUNT(SPI_COUNT)
		, .SPI_DATA_I(SPI_DATA_I)
		);

   always @(posedge CLK)
      begin
	if (SPI_DATA_VALID) 
		r_data <= SPI_DATA;
      end
    assign SPI_DATA_I[7:0] = r_data[7:0]; // latched version
    //assign SPI_DATA_I[7:0] = SPI_DATA[7:0]; // path thorugh 


//############  ISP TEST MODULE #####################################

`ifdef ISP_DEBUG_LED
  wire [3:0] ISP_STATE; 
`endif
  isp ISP (.CLK(CLK)
		, .MOSI(JTDI)
                , .SCK(JTCK2)
                , .CS(CS2)
                , .MISO(JTDO2)

`ifdef ISP_DEBUG_LED
		, .ISP_STATE(ISP_STATE)
`endif
		);

  wire CLK_INT;

`ifdef TARGET_MachXO
  OSCC osc_internal (.OSC(CLK_INT)); // 18 MHz - 26 MHz
`elsif TARGET_MachXO2
  //OSCH #(.NOM_FREQ("44.33")) osc_internal
  OSCH #(.NOM_FREQ("24.18")) osc_internal
  //OSCH #(.NOM_FREQ("10.23")) osc_internal // 10.23 MHz 
	(.OSC(CLK_INT), .STDBY(1'b0));

 `ifdef EXTOSC_PASSIVE
    wire EXTOSC_EN = ~EXTOSC;
 `else
    wire EXTOSC_EN = 1'b1;
  `endif
`endif
//############  CLOCK GENERATION #####################################
`ifdef TARGET_MachXO
    wire CLK_IN0 = CLK_INT;
`elsif TARGET_MachXO2
  DCMA clk_selector (.CLK0(CLK_INT) 
	, .CLK1(EXTOSC)
	, .SEL(`CLK_SEL) // 0: CLK0 1: CLK1
	, .DCMOUT(CLK_IN0)
	);
`endif

`ifdef SYS_DRIVEN_CLK4PH  //  SYS_DRIVEN_CLK4PH
  `ifdef USE_DMY_CLOCK    //  SYS_DRIVEN_CLK4PH (with DMY_CLOCK)
    reg [3:0] r_clk_0     = 4'b0110;
    reg [3:0] r_clk_90    = 4'b0011;
    reg [3:0] r_clk_180   = 4'b1001;
    reg [3:0] r_clk_270   = 4'b1100;
    
    always @(negedge CLK_IN0) // CLK4X
    begin
	r_clk_0   <=  { r_clk_0[2:0]   , r_clk_0[3] };
	r_clk_90  <=  { r_clk_90[2:0]  , r_clk_90[3] };
	r_clk_180 <=  { r_clk_180[2:0] , r_clk_180[3] };
	r_clk_270 <=  { r_clk_270[2:0] , r_clk_270[3] };
    end

    wire CLK0_OUT   =  r_clk_0[3];
    wire CLK90_OUT  =  r_clk_90[3];
    wire CLK180_OUT =  r_clk_180[3];
    wire CLK270_OUT =  r_clk_270[3];
    assign CLK      =  CLK0_OUT;
    assign CLK_IN   =  CLK0_OUT;

  `else                   //  SYS_DRIVEN_CLK4PH (with PLL)
	EHXPLLJ #(
		.CLKI_DIV(24)
		, .CLKFB_DIV(24)
		, .CLKOP_DIV(32)
		, .CLKOS_DIV(32)
		, .CLKOS2_DIV(32)
		, .CLKOS3_DIV(32)
		, .CLKOP_FPHASE(0)
		, .CLKOS_FPHASE(2)
		, .CLKOS2_FPHASE(4)
		, .CLKOS3_FPHASE(6)
	) pll
	(.CLKI(CLK_IN0)
		, .CLKOP(CLK0_OUT)
		, .CLKOS(CLK90_OUT)
		, .CLKOS2(CLK180_OUT)
		, .CLKOS3(CLK270_OUT)
		, .ENCLKOS(1'b1)
		, .ENCLKOS2(1'b1)
		, .ENCLKOS3(1'b1)
	);
	assign CLK    = CLK0_OUT;
  `endif
`else                     //  SYS_DRIVEN_CLK2X
	assign CLK_IN = CLK_IN0;
	assign CLK    = CLK_IN;
`endif

//############  CLOCK TEST #####################################

  reg [26:0] clk_count;
  
  always @(posedge CLK)
    begin
         clk_count = clk_count + 1;
    end

//`define CLK_TEST
`define TEST_02
`ifdef CLK_TEST
  reg [26:0] clk_count0;
  reg [26:0] clk_count1;
  
  always @(posedge CLK_IN0)
    begin
         clk_count0 = clk_count0 + 1;
    end

`ifdef SYS_DRIVEN_CLK4PH
  always @(posedge CLK0_OUT)
    begin
         clk_count1 = clk_count1 + 1;
    end
  reg [26:0] clk_count2;
  reg [26:0] clk_count3;
  reg [26:0] clk_count4;
  always @(posedge CLK90_OUT)
    begin
         clk_count2 = clk_count2 + 1;
    end
  always @(posedge CLK180_OUT)
    begin
         clk_count3 = clk_count3 + 1;
    end
  always @(posedge CLK270_OUT)
    begin
         clk_count4 = clk_count4 + 1;
    end
`else
  always @(posedge CLK_IN)
    begin
         clk_count1 = clk_count1 + 1;
    end
`endif

  assign LED[7] = ~(clk_count[24] & (clk_count[13:10] == 0)); // 1/16 duty
  assign LED[6] = CS1;
  assign LED[5] = CS2;
`ifdef SYS_DRIVEN_CLK4PH
 `ifdef ISP_DEBUG_LED
    assign LED[4] = ~ISP_STATE[3];
    assign LED[3] = ~ISP_STATE[2];
    assign LED[2] = ~ISP_STATE[1];
    assign LED[1] = ~ISP_STATE[0];
 `elsif SPI_DEBUG_LED
    assign LED[4] = ~SPI_DATA[3];
    assign LED[3] = ~SPI_DATA[2];
    assign LED[2] = ~SPI_DATA[1];
    assign LED[1] = ~SPI_DATA[0];
 `else
    assign LED[4] = ~( clk_count1[24] & (clk_count1[18:14] <= 5'h1));
    assign LED[3] = ~( clk_count2[24] & (clk_count2[18:14] <= 5'h2));
    assign LED[2] = ~( clk_count3[24] & (clk_count3[18:14] <= 5'h4));
    assign LED[1] = ~( clk_count4[24] & (clk_count4[18:14] <= 5'h8));
 `endif
`else
  assign LED[4] = ~clk_count1[24];
  assign LED[3] = 1'b1;
  assign LED[2] = 1'b1;
  assign LED[1] = 1'b1;
`endif
  assign LED[0] = 1'b1;

`elsif TEST_02
  assign LED[7] = ~( clk_count[24] & (clk_count[13:10] == 0));
  assign LED[6] = CS2 | (clk_count[13:10] != 0);
  assign LED[5] = CS1 | (clk_count[13:10] != 0);
  assign LED[4] = ~(CS2? SPI_DATA[4]: 1'b0);
  assign LED[3] = ~(CS2? SPI_DATA[3]: ISP_STATE[3]);
  assign LED[2] = ~(CS2? SPI_DATA[2]: ISP_STATE[2]);
  assign LED[1] = ~(CS2? SPI_DATA[1]: ISP_STATE[1]);
  assign LED[0] = ~(CS2? SPI_DATA[0]: ISP_STATE[0]);
`else // TEST_01
  // 1st : check led and internal clock
  assign LED[0] = ~(clk_count[26:24] == 0);
  assign LED[1] = ~(clk_count[26:24] == 1);
  assign LED[2] = ~(clk_count[26:24] == 2);
  assign LED[3] = ~(clk_count[26:24] == 3);
  assign LED[4] = ~(clk_count[26:24] == 4);
  assign LED[5] = ~(clk_count[26:24] == 5);
  assign LED[6] = ~(clk_count[26:24] == 6);
  assign LED[7] = ~(clk_count[26:24] == 7);
`endif
endmodule
